"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.authRequestWithRetry = exports.refreshTokens = exports.fetchTokens = exports.getRefreshTokenFn = exports.getPushRefreshTokenFn = exports.getOdspRefreshTokenFn = exports.getLoginPageUrl = exports.getFetchTokenUrl = exports.pushScope = exports.getOdspScope = void 0;
const internal_1 = require("@fluidframework/core-utils/internal");
const internal_2 = require("@fluidframework/telemetry-utils/internal");
const odspDocLibUtils_js_1 = require("./odspDocLibUtils.js");
const odspErrorUtils_js_1 = require("./odspErrorUtils.js");
const odspRequest_js_1 = require("./odspRequest.js");
/**
 * @alpha
 */
const getOdspScope = (server) => `offline_access ${(0, odspDocLibUtils_js_1.getSiteUrl)(server)}/AllSites.Write`;
exports.getOdspScope = getOdspScope;
/**
 * @alpha
 */
exports.pushScope = "offline_access https://pushchannel.1drv.ms/PushChannel.ReadWrite.All";
/**
 * @internal
 */
function getFetchTokenUrl(server) {
    return `${(0, odspDocLibUtils_js_1.getAadUrl)(server)}/${(0, odspDocLibUtils_js_1.getAadTenant)(server)}/oauth2/v2.0/token`;
}
exports.getFetchTokenUrl = getFetchTokenUrl;
/**
 * @internal
 */
function getLoginPageUrl(server, clientConfig, scope, odspAuthRedirectUri) {
    return (`${(0, odspDocLibUtils_js_1.getAadUrl)(server)}/${(0, odspDocLibUtils_js_1.getAadTenant)(server)}/oauth2/v2.0/authorize?` +
        `client_id=${clientConfig.clientId}` +
        `&scope=${scope}` +
        `&response_type=code` +
        `&redirect_uri=${odspAuthRedirectUri}`);
}
exports.getLoginPageUrl = getLoginPageUrl;
/**
 * @internal
 */
const getOdspRefreshTokenFn = (server, clientConfig, tokens) => (0, exports.getRefreshTokenFn)((0, exports.getOdspScope)(server), server, clientConfig, tokens);
exports.getOdspRefreshTokenFn = getOdspRefreshTokenFn;
/**
 * @internal
 */
const getPushRefreshTokenFn = (server, clientConfig, tokens) => (0, exports.getRefreshTokenFn)(exports.pushScope, server, clientConfig, tokens);
exports.getPushRefreshTokenFn = getPushRefreshTokenFn;
/**
 * @internal
 */
const getRefreshTokenFn = (scope, server, clientConfig, tokens) => async () => {
    const newTokens = await refreshTokens(server, scope, clientConfig, tokens);
    return newTokens.accessToken;
};
exports.getRefreshTokenFn = getRefreshTokenFn;
/**
 * Fetch an access token and refresh token from AAD
 * @param server - The server to auth against
 * @param scope - The desired oauth scope
 * @param clientConfig - Info about this client's identity
 * @param credentials - Credentials authorizing the client for the requested token
 * @internal
 */
async function fetchTokens(server, scope, clientConfig, credentials) {
    const body = {
        scope,
        client_id: clientConfig.clientId,
        ...credentials,
    };
    const response = await (0, odspRequest_js_1.unauthPostAsync)(getFetchTokenUrl(server), new URLSearchParams(body));
    const parsedResponse = await response.json();
    const accessToken = parsedResponse.access_token;
    const refreshToken = parsedResponse.refresh_token;
    if (accessToken === undefined || refreshToken === undefined) {
        try {
            (0, odspErrorUtils_js_1.throwOdspNetworkError)(
            // pre-0.58 error message: unableToGetAccessToken
            "Unable to get access token.", parsedResponse.error === "invalid_grant" ? 401 : response.status, response);
        }
        catch (error) {
            if ((0, internal_2.isFluidError)(error) && isAccessTokenError(parsedResponse)) {
                error.addTelemetryProperties({
                    innerError: parsedResponse.error,
                    errorDescription: parsedResponse.error_description,
                    code: JSON.stringify(parsedResponse.error_codes),
                    timestamp: parsedResponse.timestamp,
                    traceId: parsedResponse.trace_id,
                    correlationId: parsedResponse.correlation_id,
                });
            }
            throw error;
        }
    }
    return { accessToken, refreshToken };
}
exports.fetchTokens = fetchTokens;
function isAccessTokenError(parsedResponse) {
    return typeof parsedResponse?.error === "string" && Array.isArray(parsedResponse?.error_codes);
}
/**
 * Fetch fresh tokens.
 * @param server - The server to auth against
 * @param scope - The desired oauth scope
 * @param clientConfig - Info about this client's identity
 * @param tokens - The tokens object provides the refresh token for the request
 *
 * @returns The tokens object with refreshed tokens.
 * @internal
 */
async function refreshTokens(server, scope, clientConfig, tokens) {
    // Clear out the old tokens while awaiting the new tokens
    const refresh_token = tokens.refreshToken;
    (0, internal_1.assert)(refresh_token.length > 0, 0x1ec /* "No refresh token provided." */);
    const credentials = {
        grant_type: "refresh_token",
        refresh_token,
    };
    const newTokens = await fetchTokens(server, scope, clientConfig, credentials);
    // Instead of returning, update the passed in tokens object
    return { accessToken: newTokens.accessToken, refreshToken: newTokens.refreshToken };
}
exports.refreshTokens = refreshTokens;
/**
 * Issue the requestCallback, providing the proper auth header based on authRequestInfo,
 * and retrying with a refreshed token if necessary.
 * @internal
 */
async function authRequestWithRetry(authRequestInfo, requestCallback) {
    const createConfig = (token) => ({ headers: { Authorization: `Bearer ${token}` } });
    const result = await requestCallback(createConfig(authRequestInfo.accessToken));
    if (authRequestInfo.refreshTokenFn && (result.status === 401 || result.status === 403)) {
        // Unauthorized, try to refresh the token
        const refreshedAccessToken = await authRequestInfo.refreshTokenFn();
        return requestCallback(createConfig(refreshedAccessToken));
    }
    return result;
}
exports.authRequestWithRetry = authRequestWithRetry;
//# sourceMappingURL=odspAuth.js.map